use std::net::Ipv4Addr;

use fatfs::FsOptions;
use serde::Serialize;
use serde_with::skip_serializing_none;
use std::collections::HashMap;
use std::io::{prelude::*, Cursor};
use trust_dns_server::proto::rr::Name;

use nzr_api::net::{cidr::CidrV4, mac::MacAddr};

#[derive(Debug, Serialize)]
#[serde(rename_all = "kebab-case")]
pub struct Metadata<'a> {
    instance_id: &'a str,
    local_hostname: &'a str,
    public_keys: Option<Vec<&'a String>>,
}

impl<'a> Metadata<'a> {
    pub fn new(instance_id: &'a str) -> Self {
        Self {
            instance_id,
            local_hostname: instance_id,
            public_keys: None,
        }
    }

    pub fn ssh_pubkeys(mut self, pubkeys: &'a [String]) -> Self {
        self.public_keys = Some(pubkeys.iter().filter(|i| !i.is_empty()).collect());
        self
    }
}

#[derive(Debug, Serialize)]
pub struct NetworkMeta<'a> {
    version: u32,
    ethernets: HashMap<String, EtherNic<'a>>,
    #[serde(skip)]
    ethnum: u8,
}

impl<'a> NetworkMeta<'a> {
    pub fn new() -> Self {
        Self {
            version: 2,
            ethernets: HashMap::new(),
            ethnum: 0,
        }
    }

    /// Define a NIC with a static address.
    pub fn static_nic(
        mut self,
        match_data: EtherMatch<'a>,
        cidr: &'a CidrV4,
        gateway: &'a Ipv4Addr,
        dns: DNSMeta<'a>,
    ) -> Self {
        self.ethernets.insert(
            format!("eth{}", self.ethnum),
            EtherNic {
                r#match: match_data,
                addresses: Some(vec![cidr]),
                gateway4: Some(gateway),
                dhcp4: false,
                nameservers: Some(dns),
            },
        );
        self.ethnum += 1;
        self
    }

    #[allow(dead_code)]
    pub fn dhcp_nic(mut self, match_data: EtherMatch<'a>) -> Self {
        self.ethernets.insert(
            format!("eth{}", self.ethnum),
            EtherNic {
                r#match: match_data,
                addresses: None,
                gateway4: None,
                dhcp4: true,
                nameservers: None,
            },
        );
        self.ethnum += 1;
        self
    }
}

#[derive(Debug, Serialize)]
pub struct Ethernets<'a> {
    nics: Vec<EtherNic<'a>>,
}

#[derive(Debug, Serialize)]
pub struct EtherNic<'a> {
    r#match: EtherMatch<'a>,
    addresses: Option<Vec<&'a CidrV4>>,
    gateway4: Option<&'a Ipv4Addr>,
    dhcp4: bool,
    nameservers: Option<DNSMeta<'a>>,
}

#[skip_serializing_none]
#[derive(Default, Debug, Serialize)]
pub struct EtherMatch<'a> {
    name: Option<&'a str>,
    macaddress: Option<&'a MacAddr>,
    driver: Option<&'a str>,
}

impl<'a> EtherMatch<'a> {
    #[allow(dead_code)]
    pub fn name(name: &'a str) -> Self {
        Self {
            name: Some(name),
            ..Default::default()
        }
    }

    pub fn mac_addr(addr: &'a MacAddr) -> Self {
        Self {
            macaddress: Some(addr),
            ..Default::default()
        }
    }

    #[allow(dead_code)]
    pub fn driver(driver: &'a str) -> Self {
        Self {
            driver: Some(driver),
            ..Default::default()
        }
    }
}

#[derive(Debug, Serialize)]
pub struct DNSMeta<'a> {
    search: Vec<Name>,
    addresses: &'a Vec<Ipv4Addr>,
}

impl<'a> DNSMeta<'a> {
    pub fn with_addrs(search: Option<Vec<Name>>, addrs: &'a Vec<Ipv4Addr>) -> Self {
        Self {
            addresses: addrs,
            search: search.unwrap_or_default(),
        }
    }
}

pub fn create_image<B>(
    metadata: &Metadata,
    netconfig: &NetworkMeta,
    user_data: Option<&B>,
) -> Result<Cursor<Vec<u8>>, Box<dyn std::error::Error>>
where
    B: AsRef<[u8]>,
{
    let mut image: Cursor<Vec<u8>> = Cursor::new(Vec::new());

    // format a:
    fatfs::format_volume(
        &mut image,
        fatfs::FormatVolumeOptions::new()
            .volume_label(*b"cidata     ")
            .fat_type(fatfs::FatType::Fat12)
            .total_sectors(2880),
    )?;

    {
        let fs = fatfs::FileSystem::new(&mut image, FsOptions::new())?;
        let rootdir = fs.root_dir();

        let md_data = serde_yaml::to_string(&metadata)?;
        let mut md_fd = rootdir.create_file("meta-data")?;
        md_fd.write_all(md_data.as_bytes())?;

        let net_data = serde_yaml::to_string(&netconfig)?;
        let mut net_fd = rootdir.create_file("network-config")?;
        net_fd.write_all(net_data.as_bytes())?;

        // user-data MUST exist, even if there is no user-data
        let mut user_fd = rootdir.create_file("user-data")?;
        if let Some(user_data) = user_data {
            user_fd.write_all(user_data.as_ref())?;
        }
    }

    Ok(image)
}
