use tokio::sync::RwLock;
use virt::stream::Stream;

use super::*;
use crate::cloud::{DNSMeta, EtherMatch, Metadata, NetworkMeta};
use crate::ctrl::net::Subnet;
use crate::ctrl::virtxml::build::DomainBuilder;
use crate::ctrl::virtxml::{DiskDeviceType, SerialType, VolType, Volume};
use crate::ctrl::vm::{InstDb, Instance, InstanceError, Progress};
use crate::ctrl::Storable;
use crate::ctx::Context;
use crate::prelude::*;
use crate::virt::VirtVolume;
use log::*;
use nzr_api::args;
use nzr_api::net::mac::MacAddr;
use std::sync::Arc;
use trust_dns_server::proto::rr::Name;

const VIRT_MAC_OUI: &[u8] = &[0x02, 0xf1, 0x0f];

macro_rules! progress {
    ($task:ident, $pct:literal, $($arg:tt)*) => {{
        let mut pt = $task.write().await;
        pt.status_text = format!($($arg)*);
        pt.percentage = $pct;
    }};
}

/// Creates a new instance
pub async fn new_instance(
    ctx: Context,
    prog_task: Arc<RwLock<Progress>>,
    args: &args::NewInstance,
) -> Result<Instance, Box<dyn std::error::Error>> {
    progress!(prog_task, 0.0, "Starting...");
    // find the subnet corresponding to the interface
    let subnet = Subnet::get_by_key(ctx.db.clone(), args.subnet.as_bytes())
        .map_err(|er| cmd_error!("Unable to get interface: {}", er))?
        .map_or(
            Err(cmd_error!(
                "Subnet {} wasn't found in database",
                &args.subnet
            )),
            Ok,
        )?;

    // bail if a domain already exists
    if let Ok(dom) = virt::domain::Domain::lookup_by_name(&ctx.virt.conn, &args.name) {
        Err(cmd_error!(
            "Domain with name already exists (uuid {})",
            dom.get_uuid_string().unwrap_or("unknown".to_owned())
        ))
    } else {
        // make sure the base image exists
        let mut base_image = VirtVolume::lookup_by_name(&ctx.virt.pools.baseimg, &args.base_image)
            .map_err(|er| cmd_error!("Couldn't find base image: {}", er))?;
        progress!(prog_task, 10.0, "Generating metadata...");

        // generate a new lease with a new MAC addr
        let mac_addr = {
            let bytes = [VIRT_MAC_OUI, rand::random::<[u8; 3]>().as_ref()].concat();
            MacAddr::from_bytes(bytes)
        }
        .map_err(|er| cmd_error!("Unable to create a new MAC address: {}", er))?;
        let lease = subnet
            .new_lease(&mac_addr, &args.name)
            .map_err(|er| cmd_error!("Failed to generate a new lease: {}", er))?;

        // generate cloud-init data
        let meta = Metadata::new(&args.name).ssh_pubkeys(&args.ssh_keys);
        let netconfig = NetworkMeta::new().static_nic(
            EtherMatch::mac_addr(&mac_addr),
            &lease.ipv4_addr,
            &subnet.gateway4,
            DNSMeta::with_addrs(
                {
                    let mut search: Vec<Name> = vec![ctx.config.dns.default_zone.clone()];
                    if let Some(zone) = &subnet.domain_name {
                        search.push(zone.clone());
                    }
                    Some(search)
                },
                &subnet.dns,
            ),
        );
        let ci_data = crate::cloud::create_image(&meta, &netconfig, None as Option<&Vec<u8>>)
            .map_err(|er| cmd_error!("Unable to create initial cloud-init image: {}", er))?
            .into_inner();

        // and upload it to a vol
        let vol_data = Volume::new(&args.name, VolType::Raw, datasize!(1440 KiB));
        let mut cidata_vol = VirtVolume::create_xml(&ctx.virt.pools.cidata, vol_data, 0).await?;

        let cistream = Stream::new(&cidata_vol.get_connect()?, 0)?;
        if let Err(er) = cidata_vol.upload(&cistream, 0, datasize!(1440 KiB).into(), 0) {
            cistream.abort().ok();
            cidata_vol.delete(0)?;
            Err(cmd_error!("Failed to create cloud-init volume: {}", er))
        } else {
            let mut idx: usize = 0;
            while idx < ci_data.len() {
                match cistream.send(&ci_data[idx..ci_data.len()]) {
                    Ok(sz) => idx += sz,
                    Err(er) => {
                        cistream.abort().ok();
                        cidata_vol.delete(0)?;
                        return Err(cmd_error!("Failed uploading to cloud-init image: {}", er));
                    }
                }
            }
            // mark the stream as finished
            cistream.finish()?;

            progress!(prog_task, 30.0, "Creating instance images...");
            // create primary volume from base image
            let mut pri_vol = base_image
                .clone_vol(
                    &ctx.virt.pools.primary,
                    &args.name,
                    datasize!((args.disk_sizes.0) GiB),
                )
                .await
                .map_err(|er| cmd_error!("Failed to clone base image: {}", er))?;

            // and, if it exists: the second volume
            let sec_vol = match args.disk_sizes.1 {
                Some(sec_size) => {
                    let voldata = Volume::new(
                        &args.name,
                        ctx.virt.pools.secondary.xml.vol_type(),
                        datasize!(sec_size GiB),
                    );
                    Some(VirtVolume::create_xml(&ctx.virt.pools.secondary, voldata, 0).await?)
                }
                None => None,
            };

            // build domain xml
            let ifname = subnet.ifname.clone();
            let devname = format!(
                "veth-{:02x}{:02x}{:02x}",
                mac_addr[3], mac_addr[4], mac_addr[5]
            );
            progress!(prog_task, 60.0, "Initializing instance...");
            let (mut inst, conn) = Instance::new(ctx.clone(), subnet, lease, {
                let pri_name = &ctx.virt.pools.primary.xml.name;
                let sec_name = &ctx.virt.pools.secondary.xml.name;
                let cidata_name = &ctx.virt.pools.cidata.xml.name;

                let mut instdata = DomainBuilder::default()
                    .name(&args.name)
                    .memory(datasize!((args.memory) MiB))
                    .cpu_topology(1, 1, args.cores, 1)
                    .net_device(|nd| {
                        nd.mac_addr(&mac_addr)
                            .with_bridge(&ifname)
                            .target_dev(&devname)
                    })
                    .disk_device(|dsk| {
                        dsk.volume_source(pri_name, &pri_vol.name)
                            .target("vda", "virtio")
                            .qcow2()
                            .boot_order(1)
                    })
                    .disk_device(|fda| {
                        fda.volume_source(cidata_name, &cidata_vol.name)
                            .device_type(DiskDeviceType::Disk)
                            .target("hda", "ide")
                    })
                    .serial_device(SerialType::Pty);

                // add desription, if provided
                instdata = match &args.description {
                    Some(desc) => instdata.description(desc),
                    None => instdata,
                };

                // add second volume, if provided
                match &sec_vol {
                    Some(vol) => instdata.disk_device(|dsk| {
                        dsk.volume_source(sec_name, &vol.name)
                            .target("vdb", "virtio")
                            .qcow2()
                    }),
                    None => instdata,
                }
            })
            .await?;

            // not a fatal error, we can set autostart afterward
            if let Err(er) = conn.set_autostart(true) {
                warn!("Couldn't set autostart for domain: {}", er);
            }

            tokio::task::spawn_blocking(move || {
                if let Err(er) = conn.create() {
                    warn!("Domain defined, but couldn't be started! Error: {}", er);
                }
            })
            .await?;

            // set all volumes to persistent to avoid deletion
            pri_vol.persist = true;
            if let Some(mut sec_vol) = sec_vol {
                sec_vol.persist = true;
            }
            cidata_vol.persist = true;
            inst.persist();

            progress!(prog_task, 80.0, "Domain created!");
            debug!("Domain {} created!", inst.xml().name.as_str());
            Ok(inst)
        }
    }
}

pub async fn delete_instance(ctx: Context, name: String) -> Result<(), Box<dyn std::error::Error>> {
    let mut inst = Instance::lookup_by_name(ctx.clone(), &name)
        .await?
        .ok_or(cmd_error!("No such domain!"))?;

    let conn = inst.virt()?;
    if conn.is_active()? {
        conn.destroy()
            .map_err(|er| cmd_error!("Failed to destroy domain: {}", er))?;
    }

    inst.undefine().await?;

    Ok(())
}

pub fn prune_instances(ctx: &Context) -> Result<(), Box<dyn std::error::Error>> {
    for entity in InstDb::all(ctx.db.clone())? {
        let entity = entity?;
        if let Err(InstanceError::DomainNotFound(name)) =
            Instance::from_entity(ctx.clone(), entity.clone())
        {
            info!("Instance {} was invalid, deleting", name);
            if let Err(err) = entity.delete() {
                warn!("Couldn't delete {}: {}", name, err);
            }
        }
    }
    Ok(())
}
