use std::{
    marker::PhantomData,
    ops::{Deref, DerefMut},
};

use serde::{Deserialize, Serialize};

use log::*;
use std::fmt;

pub mod net;
pub mod virtxml;
pub mod vm;

#[derive(Clone)]
pub struct Entity<T>
where
    T: Storable + Serialize,
{
    inner: T,
    key: Vec<u8>,
    tree: sled::Tree,
    db: sled::Db,
    pub transient: bool,
}

impl<T> Deref for Entity<T>
where
    T: Storable,
{
    type Target = T;
    fn deref(&self) -> &Self::Target {
        &self.inner
    }
}

impl<T> DerefMut for Entity<T>
where
    T: Storable,
{
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.inner
    }
}

impl<T> Drop for Entity<T>
where
    T: Storable,
{
    fn drop(&mut self) {
        if self.transient {
            let key_str = String::from_utf8_lossy(&self.key);
            debug!("Transient flag enabled for {}, dropping!", &key_str);
            if let Err(err) = self.delete() {
                warn!("Couldn't delete {} from database: {}", &key_str, err);
            }
        }
    }
}

impl<T> Entity<T>
where
    T: Storable,
{
    pub fn transient<V>(inner: T, key: V, tree: sled::Tree, db: sled::Db) -> Self
    where
        V: AsRef<[u8]>,
    {
        Entity {
            inner,
            key: key.as_ref().to_owned(),
            tree,
            db,
            transient: true,
        }
    }

    pub fn key(&self) -> &[u8] {
        &self.key
    }
}

impl<T> Entity<T>
where
    T: Storable + Serialize,
{
    pub fn update(&self) -> Result<(), StorableError> {
        let mut bytes: Vec<u8> = Vec::new();
        ciborium::ser::into_writer(&self.inner, &mut bytes)
            .map_err(|e| StorableError::new(ErrType::SerializeFailed, e))?;
        self.tree
            .insert(&self.key, bytes.as_slice())
            .map_err(|e| StorableError::new(ErrType::DbError, e))?;
        Ok(())
    }

    pub fn replace(&mut self, other: T) -> Result<(), StorableError> {
        self.inner = other;
        self.update()
    }

    pub fn delete(&self) -> Result<(), StorableError> {
        self.on_delete(&self.db)?;
        self.tree
            .remove(&self.key)
            .map_err(|e| StorableError::new(ErrType::DbError, e))?;
        Ok(())
    }
}

#[derive(Debug)]
pub enum ErrType {
    DbError,
    DeserializeFailed,
    SerializeFailed,
}

#[derive(Debug)]
pub struct StorableError {
    err_type: ErrType,
    inner: Option<Box<dyn std::error::Error>>,
}

impl StorableError {
    fn new<E>(err_type: ErrType, inner: E) -> Self
    where
        E: std::error::Error + 'static,
    {
        Self {
            err_type,
            inner: Some(Box::new(inner)),
        }
    }
}

impl fmt::Display for ErrType {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::DbError => write!(f, "Database error"),
            Self::DeserializeFailed => write!(f, "Deserialize failed"),
            Self::SerializeFailed => write!(f, "Serialize failed"),
        }
    }
}

impl fmt::Display for StorableError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.err_type.fmt(f)?;
        if let Some(inner) = &self.inner {
            write!(f, ": {}", inner)?;
        }
        Ok(())
    }
}

impl std::error::Error for StorableError {}

pub trait Storable
where
    for<'de> Self: Deserialize<'de> + Serialize,
{
    fn tree_name() -> Option<&'static [u8]>;

    fn get_by_key(db: sled::Db, key: &[u8]) -> Result<Option<Entity<Self>>, StorableError> {
        let tree_name = match Self::tree_name() {
            Some(tn) => tn,
            None => unimplemented!(),
        };
        let tree = db
            .open_tree(tree_name)
            .map_err(|e| StorableError::new(ErrType::DbError, e))?;
        match tree
            .get(key)
            .map_err(|e| StorableError::new(ErrType::DbError, e))?
        {
            Some(vec) => {
                let deserialized: Self = ciborium::de::from_reader(&*vec)
                    .map_err(|e| StorableError::new(ErrType::DeserializeFailed, e))?;
                Ok(Some(Entity {
                    inner: deserialized,
                    key: key.to_owned(),
                    tree,
                    db,
                    transient: false,
                }))
            }
            None => Ok(None),
        }
    }

    fn insert(db: sled::Db, item: Self, key: &[u8]) -> Result<Entity<Self>, StorableError> {
        let tree_name = match Self::tree_name() {
            Some(tn) => tn,
            None => unimplemented!(),
        };
        let tree = db
            .open_tree(tree_name)
            .map_err(|e| StorableError::new(ErrType::DbError, e))?;
        let ent = Entity {
            inner: item,
            key: key.to_owned(),
            tree,
            db,
            transient: false,
        };
        ent.update()?;
        Ok(ent)
    }

    /// Requests all items from the database, as a [`StorIter`].
    fn all(db: sled::Db) -> Result<StorIter<Self>, StorableError> {
        let tree_name = match Self::tree_name() {
            Some(tn) => tn,
            None => unimplemented!(),
        };
        let tree = db
            .open_tree(tree_name)
            .map_err(|e| StorableError::new(ErrType::DbError, e))?;
        Ok(StorIter::new(db, tree))
    }

    /// Function to allow storable objects to perform actions on deletion.
    fn on_delete(&self, _db: &sled::Db) -> Result<(), StorableError> {
        // No-op
        debug!("deleting; Storable no-op!");
        Ok(())
    }
}

/// Iterator of [`Storable`]s in the running database.
pub struct StorIter<T>
where
    T: Storable,
{
    db: sled::Db,
    tree: sled::Tree,
    iter: sled::Iter,
    phantom: PhantomData<T>,
}

impl<T> StorIter<T>
where
    T: Storable,
{
    /// Creates a new iterator of [`Storable`]s using a [`sled::Db`] and
    /// [`sled::Tree`].
    fn new(db: sled::Db, tree: sled::Tree) -> Self {
        Self {
            db,
            tree: tree.clone(),
            iter: tree.iter(),
            phantom: PhantomData,
        }
    }
}

impl<T> Iterator for StorIter<T>
where
    T: Storable,
{
    type Item = Result<Entity<T>, StorableError>;

    fn next(&mut self) -> Option<Self::Item> {
        if let Some(next) = self.iter.next() {
            match next {
                Ok((key, val)) => {
                    let inner = {
                        let vec = val.to_vec();
                        let inner = ciborium::de::from_reader(vec.as_slice())
                            .map_err(|e| StorableError::new(ErrType::DeserializeFailed, e));
                        match inner {
                            Ok(inner) => inner,
                            Err(err) => {
                                return Some(Err(err));
                            }
                        }
                    };
                    Some(Ok(Entity {
                        inner,
                        key: key.to_vec(),
                        tree: self.tree.clone(),
                        db: self.db.clone(),
                        transient: false,
                    }))
                }
                Err(err) => Some(Err(StorableError::new(ErrType::DbError, err))),
            }
        } else {
            None
        }
    }
}
