pub mod build;
#[cfg(test)]
mod test;

use nzr_api::net::mac::MacAddr;
use serde::{Deserialize, Serialize};
use serde_with::skip_serializing_none;
use uuid::Uuid;

#[skip_serializing_none]
#[derive(Debug, Serialize, Deserialize, PartialEq, Eq)]
#[serde(rename = "domain")]
#[serde(rename_all = "camelCase")]
pub struct Domain {
    #[serde(rename = "@type")]
    pub r#type: String,
    pub name: String,
    pub uuid: uuid::Uuid,
    pub title: Option<String>,
    pub description: Option<String>,
    pub vcpu: Vcpu,
    pub memory: SizeInfo,
    pub current_memory: Option<SizeInfo>,
    pub features: Option<FeatureList>,
    pub cpu: Cpu,
    pub devices: DeviceList,
    pub os: OsData,
    pub on_poweroff: Option<PowerAction>,
    pub on_reboot: Option<PowerAction>,
    pub on_crash: Option<PowerAction>,
}

impl Default for Domain {
    fn default() -> Self {
        Domain {
            r#type: "kvm".to_owned(),
            name: String::default(),
            memory: SizeInfo {
                unit: SizeUnit::MiB,
                amount: 512,
            },
            current_memory: None,
            features: Some(FeatureList {
                features: vec![Feature::APIC, Feature::ACPI],
            }),
            cpu: Cpu {
                mode: "host-passthrough".to_owned(),
                topology: CpuTopology {
                    sockets: 1,
                    dies: 1,
                    cores: 2,
                    threads: 1,
                },
            },
            uuid: Uuid::new_v4(),
            title: None,
            description: None,
            vcpu: Vcpu {
                placement: "static".to_owned(),
                value: 2,
            },
            os: OsData {
                boot: Some(BootNode {
                    dev: BootDevice::HardDrive,
                }),
                r#type: OsType::default(),
                bios: BiosData {
                    useserial: "yes".to_owned(),
                    reboot_timeout: 0,
                },
            },
            on_poweroff: None,
            on_reboot: None,
            on_crash: None,
            devices: DeviceList::default(),
        }
    }
}

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct FeatureList {
    #[serde(rename = "$value")]
    features: Vec<Feature>,
}

impl std::ops::Deref for FeatureList {
    type Target = Vec<Feature>;
    fn deref(&self) -> &Self::Target {
        &self.features
    }
}

impl std::ops::DerefMut for FeatureList {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.features
    }
}

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq)]
#[serde(rename_all = "lowercase")]
#[allow(clippy::upper_case_acronyms)]
pub enum Feature {
    ACPI,
    APIC,
}

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq)]
pub enum PowerAction {
    Destroy,
    Restart,
    Preserve,
}

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct Vcpu {
    #[serde(rename = "@placement")]
    placement: String,
    #[serde(rename = "$value")]
    value: u16,
}

//  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=

#[skip_serializing_none]
#[derive(Default, Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct DiskSource {
    #[serde(rename = "@file")]
    file: Option<String>,
    #[serde(rename = "@dev")]
    dev: Option<String>,
    #[serde(rename = "@dir")]
    dir: Option<String>,
    // XXX: not supporting network type
    #[serde(rename = "@pool")]
    pub pool: Option<String>,
    #[serde(rename = "@volume")]
    pub volume: Option<String>,
}

#[skip_serializing_none]
#[derive(Default, Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct DiskDriver {
    #[serde(rename = "@name")]
    name: String,
    #[serde(rename = "@type")]
    r#type: Option<String>,
}

#[derive(Default, Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct DiskTarget {
    #[serde(rename = "@dev")]
    dev: String,
    #[serde(rename = "@bus")]
    bus: String,
}

#[derive(Clone, Copy, Default, Debug, Serialize, Deserialize, PartialEq, Eq)]
#[serde(rename_all = "lowercase")]
pub enum DiskType {
    File,
    Block,
    Volume,
    #[default]
    None,
}

#[derive(Default, Debug, Serialize, Deserialize, PartialEq, Eq)]
#[serde(rename_all = "lowercase")]
pub enum DiskDeviceType {
    Floppy,
    #[default]
    Disk,
    Cdrom,
}

#[skip_serializing_none]
#[derive(Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct DiskBoot {
    #[serde(rename = "@order")]
    order: u32,
}

#[skip_serializing_none]
#[derive(Default, Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct DiskDevice {
    #[serde(rename = "@type")]
    pub r#type: DiskType,
    boot: Option<DiskBoot>,
    #[serde(rename = "@device")]
    device: DiskDeviceType,
    pub source: DiskSource,
    target: DiskTarget,
    driver: Option<DiskDriver>,
}

//  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct NetMac {
    #[serde(rename = "@address")]
    address: MacAddr,
}

#[skip_serializing_none]
#[derive(Default, Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct NetSource {
    #[serde(rename = "@bridge")]
    bridge: Option<String>,
    #[serde(rename = "@network")]
    network: Option<String>,
}

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct NetModel {
    #[serde(rename = "@type")]
    r#type: String,
}

impl Default for NetModel {
    fn default() -> Self {
        Self {
            r#type: "virtio".to_owned(),
        }
    }
}

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct NetTarget {
    #[serde(rename = "@dev")]
    dev: String,
}

#[derive(Default, Debug, Serialize, Deserialize, PartialEq, Eq)]
#[serde(rename_all = "lowercase")]
pub enum IfaceType {
    Network,
    Bridge,
    #[default]
    None,
}

#[skip_serializing_none]
#[derive(Default, Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct NetDevice {
    #[serde(rename = "@type")]
    r#type: IfaceType,
    mac: Option<NetMac>,
    source: NetSource,
    model: NetModel,
    target: Option<NetTarget>,
}

//  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=

#[derive(Default, Debug, Serialize, Deserialize, PartialEq, Eq)]
#[serde(rename_all = "lowercase")]
pub enum SerialType {
    #[default]
    Pty,
    Unix,
    SpiceVMC,
    StdIO,
}

#[derive(Default, Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct SerialDevice {
    #[serde(rename = "@type")]
    r#type: SerialType,
}

//  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq)]
#[serde(untagged)]
pub enum Device {
    #[serde(rename = "disk")]
    Disk { disk: DiskDevice },
    #[serde(rename = "interface")]
    Interface { interface: NetDevice },
    #[serde(rename = "console")]
    Console { console: SerialDevice },
}

#[derive(Default, Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct DeviceList {
    #[serde(skip_serializing_if = "Vec::is_empty")]
    #[serde(default)]
    disk: Vec<DiskDevice>,
    #[serde(skip_serializing_if = "Vec::is_empty")]
    #[serde(default)]
    interface: Vec<NetDevice>,
    #[serde(skip_serializing_if = "Vec::is_empty")]
    #[serde(default)]
    console: Vec<SerialDevice>,
}

impl DeviceList {
    pub fn push(&mut self, dev: Device) {
        match dev {
            Device::Disk { disk } => self.disk.push(disk),
            Device::Interface { interface } => self.interface.push(interface),
            Device::Console { console } => self.console.push(console),
        }
    }

    pub fn disks(&self) -> &Vec<DiskDevice> {
        &self.disk
    }

    pub fn interfaces(&mut self) -> &Vec<NetDevice> {
        &self.interface
    }
}

//  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq)]
#[serde(rename_all = "lowercase")]
pub enum BootDevice {
    Floppy,
    Cdrom,
    #[serde(rename = "hd")]
    HardDrive,
    Network,
}

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct BootNode {
    #[serde(rename = "@dev")]
    dev: BootDevice,
}

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct BiosData {
    #[serde(rename = "@useserial")]
    useserial: String,
    #[serde(rename = "@rebootTimeout")]
    reboot_timeout: u32,
}

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct OsType {
    #[serde(rename = "@arch")]
    arch: String,
    #[serde(rename = "@machine")]
    machine: String,
    #[serde(rename = "$value")]
    value: String,
}

impl Default for OsType {
    fn default() -> Self {
        Self {
            arch: "x86_64".to_owned(),
            machine: "pc-i440fx-5.2".to_owned(),
            value: "hvm".to_owned(),
        }
    }
}

#[skip_serializing_none]
#[derive(Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct OsData {
    boot: Option<BootNode>,
    r#type: OsType,
    // we will not be doing PV, no <bootloader>/<kernel>/<initrd>/etc
    bios: BiosData,
}

impl Default for OsData {
    fn default() -> Self {
        Self {
            boot: Some(BootNode {
                dev: BootDevice::HardDrive,
            }),
            r#type: OsType::default(),
            bios: BiosData {
                useserial: "yes".to_owned(),
                reboot_timeout: 0,
            },
        }
    }
}

#[derive(Copy, Clone, Debug, Deserialize, PartialEq, Eq)]
pub enum SizeUnit {
    #[serde(rename = "bytes")]
    Bytes,
    KiB,
    MiB,
    GiB,
    TiB,
    PiB,
}

impl std::fmt::Display for SizeUnit {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                Self::Bytes => "bytes",
                Self::KiB => "KiB",
                Self::MiB => "MiB",
                Self::GiB => "GiB",
                Self::TiB => "TiB",
                Self::PiB => "PiB",
            }
        )
    }
}

impl Serialize for SizeUnit {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        serializer.serialize_str(&self.to_string())
    }
}

#[derive(Copy, Clone, Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct SizeInfo {
    #[serde(rename = "@unit")]
    pub unit: SizeUnit,
    #[serde(rename = "$value")]
    pub amount: u64,
}

impl From<SizeInfo> for u64 {
    fn from(value: SizeInfo) -> Self {
        let pow = match value.unit {
            SizeUnit::Bytes => 0,
            SizeUnit::KiB => 1,
            SizeUnit::MiB => 2,
            SizeUnit::GiB => 3,
            SizeUnit::TiB => 4,
            SizeUnit::PiB => 5,
        };
        value.amount * 1024u64.pow(pow)
    }
}

impl std::fmt::Display for SizeInfo {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{} {}", self.amount, self.unit)
    }
}

impl SizeInfo {
    pub fn qemu_style(&self) -> String {
        if self.unit == SizeUnit::Bytes {
            self.amount.to_string()
        } else {
            let unit_str = self.unit.to_string();
            format!("{}{}", self.amount, unit_str.chars().next().unwrap())
        }
    }
}

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct CpuTopology {
    #[serde(rename = "@sockets")]
    sockets: u8,
    #[serde(rename = "@dies")]
    dies: u8,
    #[serde(rename = "@cores")]
    cores: u8,
    #[serde(rename = "@threads")]
    threads: u8,
}

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct Cpu {
    #[serde(rename = "@mode")]
    mode: String,
    topology: CpuTopology,
}

//  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=

#[skip_serializing_none]
#[derive(Debug, Default, Serialize, Deserialize, PartialEq, Eq)]
#[serde(rename = "volume")]
#[serde(rename_all = "camelCase")]
pub struct Volume {
    #[serde(rename = "@type")]
    r#type: String,
    pub name: String,
    pub key: Option<String>,
    allocation: Option<SizeInfo>,
    pub capacity: Option<SizeInfo>,
    pub target: Option<VolTarget>,
}

#[derive(PartialEq, Eq)]
pub enum VolType {
    Qcow2,
    Block,
    Raw,
    Invalid,
}

impl Volume {
    pub fn new(name: &str, voltype: VolType, size: SizeInfo) -> Self {
        Self {
            name: name.to_owned(),
            r#type: match voltype {
                VolType::Qcow2 => "file",
                VolType::Block => "block",
                VolType::Raw => "file",
                VolType::Invalid => "invalid",
            }
            .to_owned(),
            capacity: Some(size),
            target: match voltype {
                VolType::Qcow2 => Some(VolTarget {
                    path: None,
                    format: Some(TargetFormat {
                        r#type: "qcow2".to_owned(),
                    }),
                }),
                VolType::Raw => Some(VolTarget {
                    path: None,
                    format: Some(TargetFormat {
                        r#type: "raw".to_owned(),
                    }),
                }),
                // block doesn't require a target aiui, libvirt is
                // able to infer
                VolType::Block => None,
                VolType::Invalid => None,
            },
            allocation: Some(size),
            key: None,
        }
    }

    pub fn vol_type(&self) -> Option<VolType> {
        if let Some(target) = &self.target {
            if let Some(format) = &target.format {
                let t = match format.r#type.as_str() {
                    "qcow2" => VolType::Qcow2,
                    "raw" => VolType::Raw,
                    "block" => VolType::Block,
                    _ => VolType::Invalid,
                };
                return Some(t);
            }
        }
        None
    }
}

#[skip_serializing_none]
#[derive(Debug, Default, Serialize, Deserialize, PartialEq, Eq)]
pub struct VolTarget {
    path: Option<String>,
    pub format: Option<TargetFormat>,
}

#[derive(Debug, Default, Serialize, Deserialize, PartialEq, Eq)]
pub struct TargetFormat {
    #[serde(rename = "@type")]
    pub r#type: String,
}

//  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=  =^..^=

#[skip_serializing_none]
#[derive(Debug, Default, Serialize, Deserialize, PartialEq, Eq)]
#[serde(rename = "pool")]
pub struct Pool {
    #[serde(rename = "@type")]
    r#type: String,
    pub name: String,
    uuid: Option<uuid::Uuid>,
    capacity: Option<SizeInfo>,
    allocation: Option<SizeInfo>,
    available: Option<SizeInfo>,
    target: Option<VolTarget>,
}

impl Pool {
    pub fn vol_type(&self) -> VolType {
        match self.r#type.as_str() {
            "fs" => VolType::Qcow2,
            "dir" => VolType::Qcow2,
            "logical" => VolType::Block,
            "zfs" => VolType::Block,
            _ => VolType::Invalid,
        }
    }
}
