use crate::ctrl::net::Lease;
use crate::ctx::Context;
use log::*;
use nzr_api::net::cidr::CidrV4;
use nzr_api::net::mac::MacAddr;
use std::net::Ipv4Addr;
use std::str::{self, Utf8Error};

use super::virtxml::{build::DomainBuilder, Domain};
use super::Storable;
use super::{net::Subnet, Entity};
use crate::virt::*;
use serde::{Deserialize, Serialize};

#[derive(Clone)]
pub struct Progress {
    pub status_text: String,
    pub percentage: f32,
}

#[derive(Clone, Serialize, Deserialize)]
pub struct InstDb {
    uuid: uuid::Uuid,
    lease_subnet: Vec<u8>,
    lease_addr: CidrV4,
}

impl InstDb {
    pub fn addr(&self) -> Ipv4Addr {
        self.lease_addr.addr
    }
}

impl Storable for InstDb {
    fn tree_name() -> Option<&'static [u8]> {
        Some(b"instances")
    }
}

impl From<Entity<InstDb>> for nzr_api::model::Instance {
    fn from(value: Entity<InstDb>) -> Self {
        nzr_api::model::Instance {
            name: String::from_utf8_lossy(&value.key).to_string(),
            uuid: value.uuid,
            lease: Some(nzr_api::model::Lease {
                subnet: String::from_utf8_lossy(&value.lease_subnet).to_string(),
                addr: value.lease_addr.clone(),
                mac_addr: MacAddr::invalid(),
            }),
            state: nzr_api::model::DomainState::NoState,
        }
    }
}

pub struct Instance {
    db_data: Entity<InstDb>,
    lease: Option<Entity<Lease>>,
    ctx: Context,
    domain_xml: Domain,
}

impl Instance {
    pub async fn new(
        ctx: Context,
        subnet: Entity<Subnet>,
        lease: Entity<Lease>,
        builder: DomainBuilder,
    ) -> Result<(Self, virt::domain::Domain), InstanceError> {
        let domain_xml = builder.build();

        let virt_domain = {
            let inst_xml =
                quick_xml::se::to_string(&domain_xml).map_err(InstanceError::CantSerialize)?;
            virt::domain::Domain::define_xml(&ctx.virt.conn, &inst_xml)
                .map_err(InstanceError::CreationFailed)?
        };

        // Get the final XML data back from libvirt; this will contain the UUID and
        // other auto-filled stuff
        let real_xml = match virt_domain.get_xml_desc(0) {
            Ok(xml_data) => match quick_xml::de::from_str::<Domain>(&xml_data) {
                Ok(xml_obj) => xml_obj,
                Err(err) => {
                    error!("Failed to deserialize XML from libvirt: {}", err);
                    if let Err(err) = virt_domain.undefine() {
                        warn!("Couldn't undefine domain after failure: {}", err);
                    }
                    return Err(InstanceError::CantDeserialize(err));
                }
            },
            Err(err) => {
                error!("Failed to get XML data from libvirt: {}", err);
                if let Err(err) = virt_domain.undefine() {
                    warn!("Couldn't undefine domain after failure: {}", err);
                }
                return Err(InstanceError::VirtError(err));
            }
        };

        debug!(
            "Adding {} (interface: {}) to the instance tree...",
            &lease.ipv4_addr, &subnet.ifname,
        );

        let db_data = InstDb {
            uuid: real_xml.uuid,
            lease_subnet: subnet.key().to_vec(),
            lease_addr: lease.ipv4_addr.clone(),
        };

        let db_data = InstDb::insert(ctx.db.clone(), db_data, real_xml.name.as_bytes())
            .map_err(InstanceError::other)?;

        let inst_obj = Instance {
            db_data,
            lease: Some(lease),
            ctx,
            domain_xml,
        };

        Ok((inst_obj, virt_domain))
    }

    pub fn uuid(&self) -> uuid::Uuid {
        self.db_data.uuid
    }

    pub fn persist(&mut self) {
        if let Some(lease) = &mut self.lease {
            lease.transient = false;
        }

        self.db_data.transient = false;
    }

    pub async fn undefine(&mut self) -> Result<(), InstanceError> {
        let virt_domain = self.virt()?;
        let connect = virt_domain
            .get_connect()
            .map_err(InstanceError::VirtError)?;

        // delete volumes
        for disk in self.domain_xml.devices.disks() {
            if let (Some(pool), Some(vol)) = (&disk.source.pool, &disk.source.volume) {
                if let Ok(vpool) = VirtPool::lookup_by_name(&connect, pool) {
                    match VirtVolume::lookup_by_name(vpool, vol) {
                        Ok(virt_vol) => {
                            if let Err(er) = virt_vol.delete(0) {
                                warn!("Can't delete {}/{}: {}", pool, vol, er);
                            }
                        }
                        Err(er) => {
                            warn!("Can't acquire handle to {}/{}: {}", pool, vol, er);
                        }
                    }
                }
            }
        }

        // undefine IP lease
        if let Some(lease) = &mut self.lease {
            lease.delete().map_err(InstanceError::other)?;
        }

        // delete instance
        virt_domain
            .undefine()
            .map_err(InstanceError::DomainDelete)?;
        self.db_data.delete().map_err(InstanceError::other)?;
        Ok(())
    }

    /// Create an Instance from a given InstDb entity.
    pub fn from_entity(ctx: Context, db_data: Entity<InstDb>) -> Result<Self, InstanceError> {
        let name = String::from_utf8_lossy(&db_data.key).into_owned();
        let virt_domain = match virt::domain::Domain::lookup_by_name(&ctx.virt.conn, &name) {
            Ok(inst) => Ok(inst),
            Err(err) => {
                if err.code() == virt::error::ErrorNumber::NoDomain {
                    // domain not found
                    Err(InstanceError::DomainNotFound(name.to_owned()))
                } else {
                    Err(InstanceError::VirtError(err))
                }
            }
        }?;
        let domain_xml: Domain = {
            let xml_str = virt_domain
                .get_xml_desc(0)
                .map_err(InstanceError::VirtError)?;
            quick_xml::de::from_str(&xml_str).map_err(InstanceError::CantDeserialize)?
        };

        let lease = match Subnet::get_by_key(ctx.db.clone(), &db_data.lease_subnet)
            .map_err(InstanceError::other)?
        {
            Some(subnet) => subnet
                .leases()
                .map_err(InstanceError::other)?
                .find(|l| {
                    if let Ok(lease) = l {
                        lease.ipv4_addr == db_data.lease_addr
                    } else {
                        false
                    }
                })
                .map(|o| o.unwrap()),
            None => None,
        };

        Ok(Self {
            ctx,
            domain_xml,
            db_data,
            lease,
        })
    }

    pub async fn lookup_by_name(ctx: Context, name: &str) -> Result<Option<Self>, InstanceError> {
        let db_data = match InstDb::get_by_key(ctx.db.clone(), name.as_bytes())
            .map_err(InstanceError::other)?
        {
            Some(data) => data,
            None => {
                return Ok(None);
            }
        };

        // TODO: handle from_instdb having None?
        Self::from_entity(ctx, db_data).map(Some)
    }

    pub fn virt(&self) -> Result<virt::domain::Domain, InstanceError> {
        let name = self.domain_xml.name.as_str();
        match virt::domain::Domain::lookup_by_name(&self.ctx.virt.conn, name) {
            Ok(inst) => Ok(inst),
            Err(err) => {
                if err.code() == virt::error::ErrorNumber::NoDomain {
                    // domain not found
                    Err(InstanceError::DomainNotFound(name.to_owned()))
                } else {
                    Err(InstanceError::VirtError(err))
                }
            }
        }
    }

    pub fn xml(&self) -> &Domain {
        &self.domain_xml
    }

    pub fn ip_lease(&self) -> Option<&Lease> {
        self.lease.as_deref()
    }
}

impl From<&Instance> for nzr_api::model::Instance {
    fn from(value: &Instance) -> Self {
        nzr_api::model::Instance {
            name: value.domain_xml.name.clone(),
            uuid: value.domain_xml.uuid,
            lease: value.lease.as_ref().map(|l| nzr_api::model::Lease {
                subnet: l.subnet.clone(),
                addr: l.ipv4_addr.clone(),
                mac_addr: l.mac_addr.clone(),
            }),
            state: value.virt().map_or(Default::default(), |domain| {
                domain
                    .get_state()
                    .map_or(Default::default(), |(code, _reason)| code.into())
            }),
        }
    }
}

#[derive(Debug)]
pub enum InstanceError {
    VirtError(virt::error::Error),
    NotInDb,
    CantDeserialize(quick_xml::de::DeError),
    CantSerialize(quick_xml::de::DeError),
    DbError(sled::Error),
    MalformedData,
    DomainNotFound(String),
    CreationFailed(virt::error::Error),
    BadInterface(Utf8Error),
    NoSubnetForInterface,
    Other(Box<dyn std::error::Error>),
    LeaseNotInDb,
    DomainDelete(virt::error::Error),
    LeaseUndefined,
}

impl std::fmt::Display for InstanceError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::VirtError(er) => er.fmt(f),
            Self::NotInDb => write!(f, "Domain exists in libvirt but is not in database"),
            Self::CantDeserialize(er) => write!(f, "Deserializing domain XML failed: {}", er),
            Self::CantSerialize(er) => write!(f, "Serializing domain XML failed: {}", er),
            Self::DbError(er) => write!(f, "Database error: {}", er),
            Self::DomainNotFound(name) => write!(f, "No domain {} found in libvirt", name),
            Self::MalformedData => write!(f, "Entry has malformed data in database"),
            Self::CreationFailed(er) => write!(f, "Error while creating domain: {}", er),
            Self::BadInterface(er) => {
                write!(f, "Couldn't get interface name from database: {}", er)
            }
            Self::NoSubnetForInterface => {
                write!(f, "Interface associated with instance isn't in database!")
            }
            Self::LeaseNotInDb => write!(
                f,
                "Found IP address, but it doesn't correspond to a lease in the database"
            ),
            Self::DomainDelete(ve) => write!(f, "Couldn't delete libvirt domain: {}", ve),
            Self::LeaseUndefined => write!(f, "Lease has been undefined by another function"),
            Self::Other(er) => er.fmt(f),
        }
    }
}

impl InstanceError {
    fn other<E>(err: E) -> Self
    where
        E: std::error::Error + 'static,
    {
        Self::Other(Box::new(err))
    }
}

impl std::error::Error for InstanceError {}
