use std::{fmt, ops::Deref};
use virt::connect::Connect;

use crate::{dns::ZoneData, virt::VirtPool};
use nzr_api::config::Config;
use std::sync::Arc;

pub struct PoolRefs {
    pub primary: VirtPool,
    pub secondary: VirtPool,
    pub cidata: VirtPool,
    pub baseimg: VirtPool,
}

impl PoolRefs {
    pub fn find_pool(&self, name: &str) -> Option<&VirtPool> {
        for pool in [&self.primary, &self.secondary, &self.baseimg, &self.cidata] {
            if let Ok(pool_name) = pool.get_name() {
                if pool_name == name {
                    return Some(pool);
                }
            }
        }
        None
    }
}

pub struct VirtCtx {
    pub conn: virt::connect::Connect,
    pub pools: PoolRefs,
}

#[derive(Clone)]
pub struct Context(Arc<InnerCtx>);

impl Deref for Context {
    type Target = InnerCtx;
    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

pub struct InnerCtx {
    pub db: sled::Db,
    pub config: Config,
    pub zones: crate::dns::ZoneData,
    pub virt: VirtCtx,
}

#[derive(Debug)]
pub enum ContextError {
    Virt(virt::error::Error),
    Db(sled::Error),
    Pool(crate::virt::PoolError),
}

impl fmt::Display for ContextError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::Virt(ve) => write!(f, "Error connecting to libvirt: {}", ve),
            Self::Db(de) => write!(f, "Error opening database: {}", de),
            Self::Pool(pe) => write!(f, "Error opening pool: {}", pe),
        }
    }
}

impl std::error::Error for ContextError {}

impl InnerCtx {
    fn new(config: Config) -> Result<Self, ContextError> {
        let zones = ZoneData::new(&config.dns);
        let conn = Connect::open(&config.libvirt_uri).map_err(ContextError::Virt)?;
        virt::error::clear_error_callback();

        let pools = PoolRefs {
            primary: VirtPool::lookup_by_name(&conn, &config.storage.primary_pool)
                .map_err(ContextError::Pool)?,
            secondary: VirtPool::lookup_by_name(&conn, &config.storage.secondary_pool)
                .map_err(ContextError::Pool)?,
            cidata: VirtPool::lookup_by_name(&conn, &config.storage.ci_image_pool)
                .map_err(ContextError::Pool)?,
            baseimg: VirtPool::lookup_by_name(&conn, &config.storage.base_image_pool)
                .map_err(ContextError::Pool)?,
        };

        Ok(Self {
            db: sled::open(&config.db_path).map_err(ContextError::Db)?,
            config,
            zones,
            virt: VirtCtx { conn, pools },
        })
    }
}

impl Context {
    pub fn new(config: Config) -> Result<Self, ContextError> {
        let inner = InnerCtx::new(config)?;
        Ok(Self(Arc::new(inner)))
    }
}
