use crate::ctrl::net::Subnet;
use log::*;
use nzr_api::config::DNSConfig;
use std::borrow::Borrow;
use std::collections::{BTreeMap, HashMap};
use std::net::Ipv4Addr;
use std::ops::Deref;
use std::str::FromStr;
use std::sync::Arc;
use tokio::sync::{Mutex, RwLock};

use trust_dns_server::authority::{AuthorityObject, Catalog};
use trust_dns_server::client::rr::{LowerName, RrKey};
use trust_dns_server::proto::rr::{rdata::soa, RData, RecordSet};
use trust_dns_server::server::{Request, RequestHandler, ResponseHandler, ResponseInfo};
use trust_dns_server::{
    proto::rr::{Name, Record},
    store::in_memory::InMemoryAuthority,
};

#[derive(Clone)]
pub struct CatalogRef(Arc<RwLock<Catalog>>);

macro_rules! make_serial {
    () => {{
        use std::time::{SystemTime, UNIX_EPOCH};

        SystemTime::now()
            .duration_since(UNIX_EPOCH)
            .expect("You know what? Fuck you. *unepochs your time*")
            .as_secs() as u32
    }};
}

#[async_trait::async_trait]
impl RequestHandler for CatalogRef {
    async fn handle_request<R>(&self, request: &Request, response_handle: R) -> ResponseInfo
    where
        R: ResponseHandler,
    {
        self.0
            .read()
            .await
            .handle_request(request, response_handle)
            .await
    }
}

#[derive(Clone)]
pub struct ZoneData(Arc<InnerZD>);

impl Deref for ZoneData {
    type Target = InnerZD;
    fn deref(&self) -> &Self::Target {
        self.0.as_ref()
    }
}

impl ZoneData {
    pub fn new(dns_config: &DNSConfig) -> Self {
        ZoneData(Arc::new(InnerZD::new(dns_config)))
    }
}

pub struct InnerZD {
    default_zone: Arc<InMemoryAuthority>,
    map: Mutex<HashMap<String, Arc<InMemoryAuthority>>>,
    catalog: CatalogRef,
    config: DNSConfig,
}

pub fn make_rectree_with_soa(name: &Name, config: &DNSConfig) -> BTreeMap<RrKey, RecordSet> {
    debug!("Creating initial SOA for {}", &name);
    let mut records: BTreeMap<RrKey, RecordSet> = BTreeMap::new();
    let soa_key = RrKey::new(
        LowerName::from(name),
        trust_dns_server::proto::rr::RecordType::SOA,
    );
    let soa_rec = Record::from_rdata(
        name.clone(),
        3600,
        RData::SOA(soa::SOA::new(
            name.clone(),
            config.soa.contact.clone(),
            make_serial!(),
            config.soa.refresh,
            config.soa.retry,
            config.soa.expire,
            3600,
        )),
    );
    records.insert(soa_key, RecordSet::from(soa_rec));
    records
}

impl InnerZD {
    pub fn new(dns_config: &DNSConfig) -> Self {
        let default_zone = Arc::new({
            let records = make_rectree_with_soa(&dns_config.default_zone, dns_config);
            InMemoryAuthority::new(
                dns_config.default_zone.clone(),
                records,
                trust_dns_server::authority::ZoneType::Primary,
                false,
            )
            .unwrap()
        });
        let mut catalog = Catalog::new();
        catalog.upsert(
            dns_config.default_zone.borrow().into(),
            Box::new(default_zone.clone()),
        );
        Self {
            default_zone,
            map: Mutex::new(HashMap::new()),
            catalog: CatalogRef(Arc::new(RwLock::new(catalog))),
            config: dns_config.clone(),
        }
    }

    pub async fn new_zone(&self, subnet: &Subnet) -> Result<(), Box<dyn std::error::Error>> {
        if let Some(name) = &subnet.domain_name {
            let auth = InMemoryAuthority::new(
                name.clone(),
                make_rectree_with_soa(name, &self.config),
                trust_dns_server::authority::ZoneType::Primary,
                false,
            )?;
            self.import(&subnet.ifname.to_string(), auth).await;
        }
        Ok(())
    }

    pub async fn import(&self, name: &str, auth: InMemoryAuthority) {
        let auth_arc = Arc::new(auth);
        log::debug!(
            "Importing {} with {} records...",
            name,
            auth_arc.records().await.len()
        );
        self.map
            .lock()
            .await
            .insert(name.to_owned(), auth_arc.clone());

        self.catalog
            .0
            .write()
            .await
            .upsert(auth_arc.origin().clone(), Box::new(auth_arc.clone()));
    }

    pub async fn delete_zone(&self, interface: &str) -> bool {
        self.map.lock().await.remove(interface).is_some()
    }

    pub async fn new_record(
        &self,
        interface: &str,
        name: &str,
        addr: Ipv4Addr,
    ) -> Result<(), Box<dyn std::error::Error>> {
        let hostname = Name::from_str(name)?;
        let zones = self.map.lock().await;
        let zone = zones.get(interface).unwrap_or(&self.default_zone);
        let fqdn = {
            let origin: Name = zone.origin().into();
            hostname.append_domain(&origin)?
        };

        log::debug!(
            "Creating new host entry {} in zone {}...",
            &fqdn,
            zone.origin()
        );

        let record = Record::from_rdata(fqdn, 3600, RData::A(addr));
        zone.upsert(record, 0).await;
        self.catalog()
            .0
            .write()
            .await
            .upsert(zone.origin().clone(), Box::new(zone.clone()));
        Ok(())
    }

    pub async fn delete_record(
        &self,
        interface: &str,
        name: &str,
    ) -> Result<bool, Box<dyn std::error::Error>> {
        let hostname = Name::from_str(name)?;
        let mut zones = self.map.lock().await;
        if let Some(zone) = zones.get_mut(interface) {
            let hostname: LowerName = hostname.into();
            self.catalog.0.write().await.remove(&hostname);
            let key = RrKey::new(hostname, trust_dns_server::proto::rr::RecordType::A);
            Ok(zone.records_mut().await.remove(&key).is_some())
        } else {
            Ok(false)
        }
    }

    pub fn catalog(&self) -> CatalogRef {
        self.catalog.clone()
    }
}
