use nzr_api::{args, model, Nazrin};
use std::borrow::Borrow;
use std::sync::Arc;
use tarpc::server::{BaseChannel, Channel};
use tarpc::tokio_serde::formats::Bincode;
use tarpc::tokio_util::codec::LengthDelimitedCodec;
use tokio::net::UnixListener;
use tokio::sync::RwLock;
use tokio::task::JoinHandle;
use uuid::Uuid;

use crate::ctrl::vm::InstDb;
use crate::ctrl::{net::Subnet, Storable};
use crate::ctx::Context;
use crate::dns::ZoneData;
use crate::{cmd, ctrl::vm::Instance};
use log::*;
use std::collections::HashMap;
use std::ops::Deref;

#[derive(Clone)]
pub struct NzrServer {
    ctx: Context,
    zones: ZoneData,
    create_tasks: Arc<RwLock<HashMap<Uuid, InstCreateStatus>>>,
}

impl NzrServer {
    pub fn new(ctx: Context, zones: ZoneData) -> Self {
        Self {
            ctx,
            zones,
            create_tasks: Arc::new(RwLock::new(HashMap::new())),
        }
    }
}

#[tarpc::server]
impl Nazrin for NzrServer {
    async fn new_instance(
        self,
        _: tarpc::context::Context,
        build_args: args::NewInstance,
    ) -> Result<uuid::Uuid, String> {
        let progress = Arc::new(RwLock::new(crate::ctrl::vm::Progress {
            status_text: "Starting...".to_owned(),
            percentage: 0.0,
        }));
        let prog_task = progress.clone();
        let build_task = tokio::spawn(async move {
            let inst = cmd::vm::new_instance(self.ctx.clone(), prog_task.clone(), &build_args)
                .await
                .map_err(|e| format!("Instance creation failed: {}", e))?;
            let addr = inst.ip_lease().map(|l| l.ipv4_addr.addr);

            {
                let mut pt = prog_task.write().await;
                pt.status_text = "Starting instance...".to_owned();
                pt.percentage = 90.0;
            }
            if let Some(addr) = addr {
                if let Err(err) = self
                    .zones
                    .new_record(&build_args.subnet, &build_args.name, addr)
                    .await
                {
                    warn!("Instance created, but no DNS record was made: {}", err);
                }
            }
            Ok((&inst).into())
        });

        let task_id = uuid::Uuid::new_v4();
        self.create_tasks.write().await.insert(
            task_id,
            InstCreateStatus {
                inner: build_task,
                progress,
            },
        );

        Ok(task_id)
    }

    async fn poll_new_instance(
        self,
        _: tarpc::context::Context,
        task_id: uuid::Uuid,
    ) -> Option<model::CreateStatus> {
        let (progress, is_finished) = {
            match self.create_tasks.read().await.get(&task_id) {
                Some(st) => (st.progress.read().await.clone(), st.inner.is_finished()),
                None => {
                    debug!("Task ID {} not found", task_id);
                    return None;
                }
            }
        };

        let result = if is_finished {
            let task = self.create_tasks.write().await.remove(&task_id).unwrap();
            Some(
                task.inner
                    .await
                    .map_err(|err| format!("Task failed with panic: {}", err))
                    .and_then(|res| res),
            )
        } else {
            None
        };

        Some(model::CreateStatus {
            status_text: progress.status_text,
            completion: progress.percentage,
            result,
        })
    }

    async fn delete_instance(self, _: tarpc::context::Context, name: String) -> Result<(), String> {
        cmd::vm::delete_instance(self.ctx.clone(), name)
            .await
            .map_err(|e| format!("Couldn't delete instance: {}", e))?;
        Ok(())
    }

    async fn get_instances(
        self,
        _: tarpc::context::Context,
        with_status: bool,
    ) -> Result<Vec<model::Instance>, String> {
        let insts: Vec<model::Instance> = InstDb::all(self.ctx.db.clone())
            .map_err(|e| e.to_string())?
            .filter_map(|i| match i {
                Ok(entity) => {
                    if with_status {
                        match Instance::from_entity(self.ctx.clone(), entity.clone()) {
                            Ok(instance) => {
                                Some(<&Instance as Into<model::Instance>>::into(&instance))
                            }
                            Err(err) => {
                                let ent_name = {
                                    let key = entity.key();
                                    String::from_utf8_lossy(key).to_string()
                                };
                                warn!("Couldn't get instance for {}: {}", err, ent_name);
                                None
                            }
                        }
                    } else {
                        Some(entity.into())
                    }
                }
                Err(err) => {
                    warn!("Iterator error: {}", err);
                    None
                }
            })
            .collect();
        Ok(insts)
    }

    async fn new_subnet(
        self,
        _: tarpc::context::Context,
        build_args: model::Subnet,
    ) -> Result<model::Subnet, String> {
        let subnet = cmd::net::add_subnet(&self.ctx, build_args)
            .await
            .map_err(|e| e.to_string())?;
        self.zones
            .new_zone(&subnet)
            .await
            .map_err(|e| e.to_string())?;
        Ok(model::Subnet {
            name: String::from_utf8_lossy(subnet.key()).to_string(),
            data: <&Subnet as Into<model::SubnetData>>::into(&subnet),
        })
    }

    async fn modify_subnet(
        self,
        _: tarpc::context::Context,
        edit_args: model::Subnet,
    ) -> Result<model::Subnet, String> {
        let subnet = Subnet::all(self.ctx.db.clone())
            .map_err(|e| e.to_string())?
            .find_map(|sub| {
                if let Ok(sub) = sub {
                    if edit_args.name.as_str() == String::from_utf8_lossy(sub.key()) {
                        Some(sub)
                    } else {
                        None
                    }
                } else {
                    None
                }
            });
        if let Some(mut subnet) = subnet {
            subnet
                .replace(edit_args.data.borrow().into())
                .map_err(|e| e.to_string())?;
            Ok(model::Subnet {
                name: edit_args.name,
                data: subnet.deref().into(),
            })
        } else {
            Err(format!("Subnet {} not found", &edit_args.name))
        }
    }

    async fn get_subnets(self, _: tarpc::context::Context) -> Result<Vec<model::Subnet>, String> {
        let subnets: Vec<model::Subnet> = Subnet::all(self.ctx.db.clone())
            .map_err(|e| e.to_string())?
            .filter_map(|s| match s {
                Ok(s) => Some(model::Subnet {
                    name: String::from_utf8(s.key().to_vec()).unwrap(),
                    data: <&Subnet as Into<model::SubnetData>>::into(s.deref()),
                }),
                Err(err) => {
                    warn!("Iterator error: {}", err);
                    None
                }
            })
            .collect();
        Ok(subnets)
    }

    async fn delete_subnet(
        self,
        _: tarpc::context::Context,
        interface: String,
    ) -> Result<(), String> {
        cmd::net::delete_subnet(&self.ctx, &interface).map_err(|e| e.to_string())?;
        self.zones.delete_zone(&interface).await;
        Ok(())
    }

    async fn garbage_collect(self, _: tarpc::context::Context) -> Result<(), String> {
        cmd::vm::prune_instances(&self.ctx).map_err(|e| e.to_string())?;
        Ok(())
    }
}

#[derive(Debug)]
pub struct GroupError(String);

impl std::fmt::Display for GroupError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "Group {} does not exist", &self.0)
    }
}

impl std::error::Error for GroupError {}

pub async fn serve(ctx: Context, zones: ZoneData) -> Result<(), Box<dyn std::error::Error>> {
    use std::os::unix::fs::PermissionsExt;

    if ctx.config.rpc.socket_path.exists() {
        std::fs::remove_file(&ctx.config.rpc.socket_path)?;
    }
    let listener = UnixListener::bind(&ctx.config.rpc.socket_path)?;

    // setup permissions so admins can actually connect to it
    std::fs::set_permissions(
        &ctx.config.rpc.socket_path,
        std::fs::Permissions::from_mode(0o770),
    )?;
    if let Some(group) = &ctx.config.rpc.admin_group {
        let group = nix::unistd::Group::from_name(group)?.ok_or(GroupError(group.clone()))?;
        nix::unistd::chown(&ctx.config.rpc.socket_path, None, Some(group.gid))?;
    }

    let codec_builder = LengthDelimitedCodec::builder();
    loop {
        debug!("Listening for new connection...");
        let (conn, _addr) = listener.accept().await?;
        let (ctx, zones) = (ctx.clone(), zones.clone());
        // hack?
        tokio::spawn(async move {
            let framed = codec_builder.new_framed(conn);
            let transport = tarpc::serde_transport::new(framed, Bincode::default());
            BaseChannel::with_defaults(transport)
                .execute(NzrServer::new(ctx, zones).serve())
                .await;
        });
    }
}

struct InstCreateStatus {
    inner: JoinHandle<Result<model::Instance, String>>,
    progress: Arc<RwLock<crate::ctrl::vm::Progress>>,
}
