use std::io::{prelude::*, BufReader};
use std::{fmt::Display, ops::Deref};
use virt::{storage_pool::StoragePool, storage_vol::StorageVol, stream::Stream};

use crate::ctrl::virtxml::VolType;
use crate::img;
use crate::{
    ctrl::virtxml::{Pool, SizeInfo, Volume},
    prelude::*,
};

use log::*;

pub struct VirtVolume {
    inner: StorageVol,
    pub persist: bool,
    pub name: String,
}

impl VirtVolume {
    fn upload_img(from: &std::fs::File, to: Stream) -> Result<(), PoolError> {
        let buf_cap: u64 = datasize!(4 MiB).into();

        let mut reader = BufReader::with_capacity(buf_cap as usize, from);
        loop {
            let read_bytes = {
                // read from the source file...
                let data = match reader.fill_buf() {
                    Ok(buf) => buf,
                    Err(er) => {
                        if let Err(er) = to.abort() {
                            warn!("Stream abort failed: {}", er);
                        }
                        return Err(PoolError::FileError(er));
                    }
                };

                if data.is_empty() {
                    break;
                }

                debug!("pulled {} bytes", data.len());

                // ... and then send upstream
                let mut send_idx = 0;
                while send_idx < data.len() {
                    match to.send(&data[send_idx..]) {
                        Ok(sz) => {
                            send_idx += sz;
                        }
                        Err(er) => {
                            if let Err(er) = to.abort() {
                                warn!("Stream abort failed: {}", er);
                            }
                            return Err(PoolError::UploadError(er));
                        }
                    }
                }
                data.len()
            };

            debug!("consuming {} bytes", read_bytes);
            reader.consume(read_bytes);
        }
        Ok(())
    }

    pub async fn create_xml(
        pool: &StoragePool,
        xmldata: Volume,
        flags: u32,
    ) -> Result<Self, Box<dyn std::error::Error>> {
        let xml = quick_xml::se::to_string(&xmldata)?;

        let svol = StorageVol::create_xml(pool, &xml, flags)?;

        if xmldata.vol_type() == Some(VolType::Qcow2) {
            let size = xmldata.capacity.unwrap().clone();
            let src_img = img::create_qcow2(size).await?;

            let stream = match Stream::new(&svol.get_connect().map_err(PoolError::VirtError)?, 0) {
                Ok(s) => s,
                Err(er) => {
                    svol.delete(0).ok();
                    return Err(Box::new(er));
                }
            };

            let img_size = src_img.metadata().unwrap().len();

            if let Err(er) = svol.upload(&stream, 0, img_size, 0) {
                svol.delete(0).ok();
                return Err(Box::new(PoolError::CantUpload(er)));
            }

            Self::upload_img(&*src_img, stream)?;
        }

        Ok(Self {
            inner: svol,
            persist: false,
            name: xmldata.name,
        })
    }

    pub fn lookup_by_name<P>(pool: P, name: &str) -> Result<Self, virt::error::Error>
    where
        P: AsRef<StoragePool>,
    {
        Ok(Self {
            inner: StorageVol::lookup_by_name(pool.as_ref(), name)?,
            // default to persisting when looking up by name
            persist: true,
            name: name.to_owned(),
        })
    }

    pub async fn clone_vol(
        &mut self,
        pool: &VirtPool,
        vol_name: &str,
        size: SizeInfo,
    ) -> Result<Self, PoolError> {
        debug!("Cloning volume to {} ({})", vol_name, &size);

        let src_path = self.get_path().map_err(PoolError::NoPath)?;

        let src_img = img::clone_qcow2(src_path, size.clone())
            .await
            .map_err(PoolError::QemuError)?;

        let newvol = Volume::new(vol_name, pool.xml.vol_type(), size);
        let newxml_str = quick_xml::se::to_string(&newvol).map_err(PoolError::SerdeError)?;
        debug!("Creating new vol...");
        let cloned = StorageVol::create_xml(pool, &newxml_str, 0).map_err(PoolError::VirtError)?;

        match cloned.get_info() {
            Ok(info) => {
                if info.capacity != u64::from(size) {
                    debug!(
                        "libvirt set wrong size {}, trying this again...",
                        info.capacity
                    );
                    if let Err(er) = cloned.resize(size.into(), 0) {
                        if let Err(er) = cloned.delete(0) {
                            warn!("Resizing disk failed, and couldn't clean up: {}", er);
                        }
                        return Err(PoolError::VirtError(er));
                    }
                } else {
                    debug!(
                        "capacity is correct ({} bytes), allocation = {} bytes",
                        info.capacity, info.allocation,
                    );
                }
            }
            Err(er) => {
                if let Err(er) = cloned.delete(0) {
                    warn!("Couldn't clean up destination volume: {}", er);
                }
                return Err(PoolError::VirtError(er));
            }
        }

        let stream = match Stream::new(&cloned.get_connect().map_err(PoolError::VirtError)?, 0) {
            Ok(s) => s,
            Err(er) => {
                cloned.delete(0).ok();
                return Err(PoolError::VirtError(er));
            }
        };

        let img_size = src_img.metadata().unwrap().len();

        if let Err(er) = cloned.upload(&stream, 0, img_size, 0) {
            cloned.delete(0).ok();
            return Err(PoolError::CantUpload(er));
        }

        Self::upload_img(&*src_img, stream)?;

        Ok(Self {
            inner: cloned,
            persist: false,
            name: vol_name.to_owned(),
        })
    }
}

impl Deref for VirtVolume {
    type Target = StorageVol;
    fn deref(&self) -> &Self::Target {
        &self.inner
    }
}

impl Drop for VirtVolume {
    fn drop(&mut self) {
        if !self.persist {
            debug!("Deleting volume {}", &self.name);
            self.inner.delete(0).ok();
        }
    }
}

#[derive(Debug)]
pub enum PoolError {
    VirtError(virt::error::Error),
    SerdeError(quick_xml::de::DeError),
    NoPath(virt::error::Error),
    FileError(std::io::Error),
    CantUpload(virt::error::Error),
    UploadError(virt::error::Error),
    QemuError(img::ImgError),
}

impl Display for PoolError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::VirtError(er) => er.fmt(f),
            Self::SerdeError(er) => er.fmt(f),
            Self::NoPath(er) => write!(f, "Couldn't get source image path: {}", er),
            Self::FileError(er) => er.fmt(f),
            Self::CantUpload(er) => write!(f, "Unable to start upload to image: {}", er),
            Self::UploadError(er) => write!(f, "Failed to upload image: {}", er),
            Self::QemuError(er) => er.fmt(f),
        }
    }
}

impl std::error::Error for PoolError {}

pub struct VirtPool {
    inner: StoragePool,
    pub xml: Pool,
}

impl Deref for VirtPool {
    type Target = StoragePool;
    fn deref(&self) -> &Self::Target {
        &self.inner
    }
}

impl AsRef<StoragePool> for VirtPool {
    fn as_ref(&self) -> &StoragePool {
        &self.inner
    }
}

impl VirtPool {
    pub fn lookup_by_name(conn: &virt::connect::Connect, id: &str) -> Result<Self, PoolError> {
        let inner = StoragePool::lookup_by_name(conn, id).map_err(PoolError::VirtError)?;
        if !inner.is_active().map_err(PoolError::VirtError)? {
            inner.create(0).map_err(PoolError::VirtError)?;
        }
        let xml_str = inner.get_xml_desc(0).map_err(PoolError::VirtError)?;
        let xml = quick_xml::de::from_str(&xml_str).map_err(PoolError::SerdeError)?;
        Ok(Self { inner, xml })
    }
}
